import os
from enma.application.core.interfaces.saver_adapter import File, ISaverAdapter


class LocalStorage(ISaverAdapter):
    def save(self, path: str, file: File) -> bool:
        os.makedirs(path, exist_ok=True)
        
        with open(os.path.join(path, f'{file.name}'), 'wb') as f:
            f.write(file.data.getbuffer())
            f.close()

        return True



from typing import Any, Literal, TypedDict, Union, TypeVar


class Title(TypedDict):
    en: str

class MangaDesc(TypedDict):
    en: str

class TagAttrs(TypedDict):
    name: dict[str, str]
    description: dict[str, str]
    group: Union[Literal["theme"], Literal["genre"], 
                 Literal["format"]]
    version: int

class IMangaTag(TypedDict):
    id: str
    type: str
    attributes: TagAttrs
    relationships: list[Any]

IAltTitles = list[dict[str, str]]

class MangaAttrs(TypedDict):
    title: Title
    altTitles: IAltTitles
    description: MangaDesc
    isLocked: bool
    links: dict[str, str]
    originalLanguage: str
    lastVolume: str
    lastChapter: str
    publicationDemographic: Any
    status: str
    year: int
    contentRating: str
    tags: list[IMangaTag]
    state: str
    chapterNumbersResetOnNewVolume: bool
    createdAt: str
    updatedAt: str
    version: int
    availableTranslatedLanguages: list[str]
    latestUploadedChapter: str

class CoverAttrs(TypedDict):
    description: str
    volume: int
    fileName: str
    locale: str
    createdAt: str
    updatedAt: str
    version: int

class CoverArtRelation(TypedDict):
    id: str
    type: Literal["cover_art"]
    attributes: CoverAttrs

class PersonAttrs(TypedDict):
    name: str
    imageUrl: str
    biography: dict[str, str]
    twitter: str
    pixiv: str
    melonBook: str
    fanBox: str
    booth: str
    namicomi: str
    nicoVideo: str
    skeb: str
    fantia: str
    tumblr: str
    youtube: str
    weibo: str
    naver: str
    website: str
    createdAt: str
    updatedAt: str
    version: int

class AuthorRelation(TypedDict):
    id: str
    type: Union[Literal["author"], Literal["artist"]]
    attributes: PersonAttrs

IRelations = list[Union[CoverArtRelation,
                       AuthorRelation,
                       dict[str, str]]]

class IManga(TypedDict):
    id: str
    type: str
    attributes: MangaAttrs
    relationships: IRelations

class IGetResult(TypedDict):
    result: str
    response: str
    data: IManga

class ISearchResult(IGetResult):
    limit: int
    offset: int
    total: int
    result: str
    response: str
    data: list[IManga]

class IChapter(TypedDict):
    chapter: str
    id: str
    others: list
    count: int

class IVolume(TypedDict):
    volume: str
    count: int
    chapters: dict[str, IChapter]

class IVolumesResponse(TypedDict):
    result: str
    volumes: dict[str, IVolume]

class IChapterHash(TypedDict):
    hash: str
    data: list[str]
    dataSaver: list[str]

class IHash(TypedDict):
    result: str
    baseUrl: str
    chapter: IChapterHash



"""
This module initializes the entrypoints library for the Enma application.
It sets up the necessary configurations and imports required for the entrypoints.
"""
from enum import Enum
from typing import Any, Generic, Optional, TypeVar, TypedDict, Union

from enma.application.core.handlers.error import InstanceError, InvalidResource, SourceNotAvailable, SourceWasNotDefined
from enma.application.core.interfaces.downloader_adapter import IDownloaderAdapter
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.saver_adapter import ISaverAdapter
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.use_cases.download_chapter import DownloadChapterRequestDTO, DownloadChapterResponseDTO, DownloadChapterUseCase, Threaded
from enma.application.use_cases.fetch_chapter_by_symbolic_link import FetchChapterBySymbolicLinkRequestDTO, FetchChapterBySymbolicLinkResponseDTO, FetchChapterBySymbolicLinkUseCase
from enma.application.use_cases.get_author_page import GetAuthorPageRequestDTO, GetAuthorPageResponseDTO, GetAuthorPageUseCase
from enma.application.use_cases.get_manga import GetMangaRequestDTO, GetMangaResponseDTO, GetMangaUseCase
from enma.application.use_cases.get_random import RandomResponseDTO, RandomUseCase
from enma.application.use_cases.paginate import PaginateRequestDTO, PaginateResponseDTO, PaginateUseCase
from enma.application.use_cases.search_manga import SearchMangaRequestDTO, SearchMangaResponseDTO, SearchMangaUseCase
from enma.domain.entities.author_page import AuthorPage
from enma.domain.entities.manga import Chapter, Manga
from enma.domain.entities.pagination import Pagination
from enma.domain.entities.search_result import SearchResult
from enma.infra.adapters.repositories.mangadex import Mangadex
from enma.infra.adapters.repositories.manganato import Manganato
from enma.infra.adapters.repositories.nhentai import NHentai, CloudFlareConfig
from enma.infra.core.interfaces.lib import IEnma
from enma.application.core.utils.logger import logger

class Sources(Enum):
    NHENTAI = 'nhentai'
    MANGADEX = 'mangadex'
    MANGANATO = 'manganato'

class ExtraConfigs(TypedDict):
    cloudflare_config: CloudFlareConfig

Source = TypeVar('Source', Sources, str)

class SourceManager(Generic[Source]):
    """
    Manages manga source repositories available to the Enma application, allowing for dynamic source selection at runtime.

    Attributes:
        source (Union[IMangaRepository, None]): The currently selected manga repository source.
        source_name (str): The name of the currently selected source.
    """

    def __init__(self) -> None:
        """
        Initializes the SourceManager with empty sources and no selected source.
        """
        self.__SOURCES: dict[str, IMangaRepository] = {}
        self.__CURRENT_SOURCE: Union[IMangaRepository, None] = None
        self.source_name = ''
    
    @property
    def source(self) -> Union[IMangaRepository, None]:
        return self.__CURRENT_SOURCE

    def get_source(self,
                   source_name: Union[Sources, Source]) -> IMangaRepository:
        """
        Retrieves a source repository by name.

        Args:
            source_name (Union[Sources, str]): The name of the source to retrieve, either as a string or an enum.

        Returns:
            IMangaRepository: The manga repository source.

        Raises:
            SourceNotAvailable: If the requested source is not available.
        """

        source_name = source_name.value if isinstance(source_name, Enum) else source_name
        source = self.__SOURCES.get(source_name)

        if source is None:
            raise SourceNotAvailable(f'{source_name} is not an available source.\nAvailable Sources {self.__SOURCES.keys()}')
        
        return source
    
    def set_source(self,
                   source_name: Union[Sources, Source]) -> None:
        """
        Sets the currently active source to the specified source name.

        Args:
            source_name (Union[Source, str]): The name of the source to activate, either as a string or an enum.
        """
        source = self.get_source(source_name=source_name)
        self.__CURRENT_SOURCE = source
        self.source_name = source_name

    def add_source(self,
                   source_name: Union[str, Sources, Source],
                   source: IMangaRepository) -> None:
        """
        Adds a new source repository to the available sources.

        Args:
            source_name (Union[str, Sources]): The name of the source to add.
            source (IMangaRepository): The manga repository source instance.

        Raises:
            InstanceError: If the provided source is not an instance of IMangaRepository.
        """
        if not isinstance(source, IMangaRepository):
            raise InstanceError('Provided source is not an instance of IMangaRepository.')
        
        self.__SOURCES[source_name if isinstance(source_name, str) else source_name.value] = source

    def remove_source(self,
                      source_name: Union[str, Source]) -> bool:

        name = source_name if isinstance(source_name, str) else source_name.value

        if name not in self.__SOURCES: 
            logger.warning(f"Source {name} not found.")
            return False
        
        del self.__SOURCES[name]
        return True
    
    def clear_sources(self) -> None:
        self.__SOURCES = dict()

def instantiate_source(callable):
    """
    Decorator function to ensure the current use case is instantiated with the current source.
    This is used to decorate methods of the Enma class that require a source to have been set.

    Args:
        callable: The method to be decorated.
    
    Returns:
        The wrapped method with source initialization logic.
    """
    def wrapper(self, *args, **kwargs):
        if self.source_manager.source is not None and \
            self._Enma__current_source_name != self.source_manager.source_name:
            self._Enma__initialize_use_case(source=self.source_manager.source)
            self._Enma__current_source_name = self.source_manager.source_name
        
        return callable(self, *args, **kwargs)
    return wrapper

class Enma(IEnma, Generic[Source]):
    """
    Main application class for Enma, providing interfaces to execute various manga-related use cases.
    Allows dynamic selection of manga sources and performs actions like fetching manga, searching, and downloading chapters.

    Attributes:
        source_manager (SourceManager[Source]): Manages the available sources and the current source selection.
    """
    def __init__(self, 
                 source: Optional[Source] = None, 
                 **kwargs) -> None:
        """
        Initializes the Enma application with optional default source selection and extra configurations.

        Args:
            source (Optional[Source], optional): The default source to be used. If provided, use cases will be initialized with this source.
        """
        self.__get_manga_use_case: Optional[IUseCase[GetMangaRequestDTO, GetMangaResponseDTO]] = None
        self.__search_manga_use_case: Optional[IUseCase[SearchMangaRequestDTO, SearchMangaResponseDTO]] = None
        self.__paginate_use_case: Optional[IUseCase[PaginateRequestDTO, PaginateResponseDTO]] = None
        self.__random_use_case: Optional[IUseCase[Any, RandomResponseDTO]] = None
        self.__downloader_use_case: Optional[IUseCase[DownloadChapterRequestDTO, DownloadChapterResponseDTO]] = None
        self.__get_author_page_use_case: Optional[IUseCase[GetAuthorPageRequestDTO, GetAuthorPageResponseDTO]] = None
        self.__fetch_chapter_by_symbolic_link_use_case: Optional[IUseCase[FetchChapterBySymbolicLinkRequestDTO, FetchChapterBySymbolicLinkResponseDTO]] = None
        self.__current_source_name: Optional[str] = None

        self.source_manager = SourceManager[Source](**kwargs)
        self.__create_default_sources()
        if source is not None: self.__initialize_use_case(source=self.source_manager.get_source(source_name=source))

    def __create_default_sources(self) -> None:
        """
        Creates and adds the default manga sources to the source manager. Currently, NHentai, Manganato, and Mangadex are added.
        """
        self.source_manager.add_source(Sources.NHENTAI, NHentai())
        self.source_manager.add_source(Sources.MANGANATO, Manganato())
        self.source_manager.add_source(Sources.MANGADEX, Mangadex())

    def __initialize_use_case(self, source: IMangaRepository) -> None:
        """
        Initializes the use cases with the given source repository. This method sets up all use cases available in Enma.

        Args:
            source (IMangaRepository): The source repository to initialize use cases with.
        """
        self.__get_manga_use_case = GetMangaUseCase(manga_repository=source)
        self.__search_manga_use_case = SearchMangaUseCase(manga_repository=source)     
        self.__paginate_use_case = PaginateUseCase(manga_repository=source)     
        self.__random_use_case = RandomUseCase(manga_repository=source)
        self.__downloader_use_case = DownloadChapterUseCase()  
        self.__get_author_page_use_case = GetAuthorPageUseCase(manga_repository=source)
        self.__fetch_chapter_by_symbolic_link_use_case = FetchChapterBySymbolicLinkUseCase(manga_repository=source)
    

    @instantiate_source
    def get(self, 
            identifier: str,
            with_symbolic_links: bool = True) -> Union[Manga, None]:
        """
        Retrieves detailed information for a specific manga identified by its ID.

        Args:
            identifier (str): The unique identifier of the manga to retrieve.
            with_symbolic_links (bool, optional): If True, fetches the manga with symbolic links to chapters. Defaults to True.

        Returns:
            Union[Manga, None]: The Manga object if found, None otherwise.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
        """
        if self.__get_manga_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')

        response = self.__get_manga_use_case.execute(dto=DTO(data=GetMangaRequestDTO(identifier=identifier,
                                                                                     with_symbolic_links=with_symbolic_links)))
        
        if not response.found: return
        return response.manga
    
    @instantiate_source
    def search(self, 
               query: str, 
               page: int=1, 
               **kwargs) -> SearchResult:
        """
        Searches for manga that match the given query string.

        Args:
            query (str): The search query string.
            page (int, optional): The page number of the search results to retrieve. Defaults to 1.
            **kwargs: Additional parameters for search customization.

        Returns:
            SearchResult: An object containing the paginated search results, including manga thumbnails.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
        """
        if self.__search_manga_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        response = self.__search_manga_use_case.execute(dto=DTO(data=SearchMangaRequestDTO(query=query,
                                                                                           page=page,
                                                                                           extra=kwargs)))
        
        return response.result
    
    @instantiate_source
    def paginate(self, 
                 page: int=1) -> Pagination:
        """
        Retrieves a specific page of manga listings.

        Args:
            page (int): The page number of manga listings to retrieve.

        Returns:
            Pagination: An object containing the paginated list of manga thumbnails and pagination details.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
        """
        if self.__paginate_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        response = self.__paginate_use_case.execute(dto=DTO(data=PaginateRequestDTO(page=page)))

        return response.result
    
    @instantiate_source
    def random(self) -> Manga:
        """
        Fetches a random manga from the currently selected source.

        Returns:
            Manga: A Manga object for the randomly selected manga.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
            NotImplementedError: If the current source does not support fetching an author's page.
        """
        if self.__random_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        response = self.__random_use_case.execute() 
        return response.result
    
    
    @instantiate_source
    def download_chapter(self, 
                         path: str, 
                         chapter: Chapter,
                         downloader: IDownloaderAdapter,
                         saver: ISaverAdapter,
                         threaded: Optional[Threaded] = None) -> None:
        """
        Downloads a manga chapter to the specified path using the provided downloader and saver adapters.

        Args:
            path (str): The filesystem path where the chapter should be saved.
            chapter (Chapter): The manga chapter to download.
            downloader (IDownloaderAdapter): The adapter to use for downloading the chapter pages.
            saver (ISaverAdapter): The adapter to use for saving the downloaded pages.
            threaded (Threaded): Determines whether the download should be performed in a threaded manner for concurrency.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
        """
        if self.__downloader_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        self.__downloader_use_case.execute(dto=DTO(data=DownloadChapterRequestDTO(chapter=chapter,
                                                                                  path=path,
                                                                                  saver_adapter=saver,
                                                                                  downloader=downloader,
                                                                                  threaded=threaded)))
    
    @instantiate_source
    def author_page(self, 
                    author: str, 
                    page: int=1) -> AuthorPage:
        """
        Fetches manga authored by a specific author.

        Args:
            author (str): The name or identifier of the author.
            page (int): The page number of results to retrieve.

        Returns:
            AuthorPage: An object containing a list of manga by the specified author.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
            NotImplementedError: If the current source does not support fetching an author's page.
        """
        if self.__get_author_page_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        return self.__get_author_page_use_case.execute(dto=DTO(data=GetAuthorPageRequestDTO(author=author,
                                                                                            page=page))).result

    @instantiate_source
    def fetch_chapter_by_symbolic_link(self, 
                                       chapter: Chapter) -> Chapter:
        """
        Fetches a manga chapter's details including pages and images by its symbolic link.

        Args:
            chapter (Chapter): The manga chapter to fetch, which must include a valid symbolic link.

        Returns:
            Chapter: An object containing the fetched chapter details such as pages and images.

        Raises:
            SourceWasNotDefined: If no source has been defined prior to calling this method.
            InvalidResource: If the provided chapter does not have a valid symbolic link.
        """
        if self.__fetch_chapter_by_symbolic_link_use_case is None:
            raise SourceWasNotDefined('You must define a source before of performing actions.')
        
        if chapter.link is None or chapter.link.link is None:
            raise InvalidResource('Chapter does not have a symbolic link.')
        
        response = self.__fetch_chapter_by_symbolic_link_use_case.execute(dto=DTO(data=FetchChapterBySymbolicLinkRequestDTO(link=chapter.link)))
        
        response.chapter.id = chapter.id

        return response.chapter



from dataclasses import dataclass
from typing import Union

from pydantic import BaseModel, Field, field_validator
from enma.application.core.handlers.error import InvalidRequest
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.search_result import SearchResult

class SearchMangaRequestDTO(BaseModel):
    query: str
    page: int = Field(default=1)
    extra: dict[str, Union[str, int]] = Field(default_factory=dict)

    @field_validator("page")
    def validate_page(cls, page: int) -> int:
        if page <= 0:
            raise InvalidRequest(message='Page value must be greater than 0.')
        return int(page)

@dataclass
class SearchMangaResponseDTO:
    result: SearchResult
    
class SearchMangaUseCase(IUseCase[SearchMangaRequestDTO, SearchMangaResponseDTO]):

    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self, dto: DTO[SearchMangaRequestDTO]) -> SearchMangaResponseDTO:
        logger.info(f'Searching for {dto.data.query}.')
        result = self.__manga_repository.search(query=dto.data.query,
                                                page=dto.data.page,
                                                **dto.data.extra)
        
        return SearchMangaResponseDTO(result=result)



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Literal, TypedDict, Union

from enma.domain.entities.base import Entity

class MIME(Enum):
    JPG = 'jpg'
    J = 'jpg'
    PNG = 'png'
    P = 'png'
    GIF = 'gif'
    G = 'gif'
    W = 'webp'
    WEBP = 'webp'
    
@dataclass
class Image:
    uri: str
    name: str = field(default='image.jpg')
    width: int = field(default=0)
    height: int = field(default=0)
    mime: MIME = field(default=MIME.J)

@dataclass
class Title:
    english: str = field(default='')
    japanese: str = field(default='')
    other: str = field(default='')

class IMangaProps(TypedDict):
    id: Union[int, str]
    created_at: datetime
    updated_at: datetime
    title: Title
    url: str
    pages_count: int
    pages: list[Image]

@dataclass
class SymbolicLink:
    link: str

@dataclass
class Chapter:
    id: Union[int, str] = field(default=0)
    pages: list[Image] = field(default_factory=list)
    pages_count: int = field(default=0)
    link: Union[SymbolicLink, None] = field(default=None) 

    def __post_init__(self) -> None:
        self.pages_count = len(self.pages)

    def add_page(self, page: Image) -> None:
        self.pages.append(page)
        self.pages_count += 1

@dataclass
class Unit:
    name: str

@dataclass
class Tag(Unit):
    type: Union[Literal["character"], 
                Literal["related"], 
                Literal["category"]]
    id: Union[int, str] = field(default=0)

@dataclass
class Author(Unit):
    id: Union[int, str] = field(default=0)

@dataclass
class Genre(Unit):
    id: Union[int, str] = field(default=0)

class ILanguage(TypedDict):
    ja: Literal['japanese']
    jp: Literal['japanese']
    japanese: Literal['japanese']
    portuguese: Literal['portuguese']
    pt: Literal['portuguese']
    pt_br: Literal['portuguese']
    english: Literal['english']
    en: Literal['english']
    en_us: Literal['english']
    chinese: Literal['chinese']
    cn: Literal['chinese']
    zh: Literal['chinese']
    russian: Literal['russian']
    ru: Literal['russian']
    turkish: Literal['turkish']
    tr: Literal['turkish']
    spanish: Literal['spanish']
    es_la: Literal['spanish']
    malay: Literal['malay']
    ms: Literal['malay']
    korean: Literal['korean']
    ko: Literal['korean']

Language: ILanguage = {
    'ja': 'japanese',
    'jp': 'japanese',
    'japanese': 'japanese',
    'portuguese': 'portuguese',
    'pt': 'portuguese',
    'pt_br': 'portuguese',
    'english': 'english',
    'en': 'english',
    'en_us': 'english',
    'chinese': 'chinese',
    'cn': 'chinese',
    'zh': 'chinese',
    'ru': 'russian',
    'russian': 'russian',
    'turkish': 'turkish',
    'tr': 'turkish',
    'spanish': 'spanish',
    'es_la': 'spanish',
    'malay': 'malay',
    'ms': 'malay',
    'korean': 'korean',
    'ko': 'korean'
}
    
class Manga(Entity[IMangaProps]):
    def __init__(self,
                 title: Title,
                 status: Literal['ongoing', 'completed'],
                 url: str,
                 chapters: Union[list[Chapter], None] = None,
                 language: Union[str, None] = None,
                 genres: Union[list[Genre], None] = None,
                 tags: Union[list[Tag], None] = None,
                 authors: Union[list[Author], None] = None,
                 thumbnail: Union[Image, None] = None,
                 cover: Union[Image, None] = None,
                 id: Union[int, str, None] = None, 
                 created_at: Union[datetime, None] = None, 
                 updated_at: Union[datetime, None] = None):
        
        super().__init__(id=id,
                         created_at=created_at,
                         updated_at=updated_at)
        
        self.status = status
        self.title = title
        self.language = language
        self.cover = cover
        self.url = url
        self.thumbnail = thumbnail
        self.authors = authors or []
        self.genres = genres or []
        self.chapters = chapters or []
        self.tags = tags or []

        self.chapters_count = len(self.chapters if self.chapters else [])

    def add_chapter(self,
                    chapter: Chapter):
        self.chapters.append(chapter)
        self.chapters_count += 1



import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

import time
from enma.infra.core.utils.cache import Cache

def test_cache_decorator():
    # Create an instance of the Cache class
    cache = Cache(max_age_seconds=10, max_size=5)

    class Fibonacci:
        @cache.cache
        def fibonacci(self, n):
            if n <= 1:
                return n
            else:
                return self.fibonacci(n-1) + self.fibonacci(n-2)

    fib = Fibonacci()
    # Call the fibonacci function multiple times with the same arguments
    result1 = fib.fibonacci(5)
    result2 = fib.fibonacci(5)
    result3 = fib.fibonacci(5)

    # Assert that the function is only executed once and the cached result is returned
    assert result1 == result2 == result3

    assert cache._CACHE.get(str([5])) == result1

    # Call the fibonacci function with a different argument
    result4 = fib.fibonacci(10)

    # Assert that a new execution is performed and the result is cached
    assert result4 == 55
    assert cache._CACHE.get(str([10])) == result4

def test_cache_expiration():
    # Create an instance of the Cache class with a short expiration time
    cache = Cache(max_age_seconds=2, max_size=5)

    @cache.cache
    def get_current_time():
        return time.time()

    # Call the function to cache its result
    result1 = get_current_time()

    # Wait for the cache entry to expire
    time.sleep(3)

    # Call the function again and expect a different result due to cache expiration
    result2 = get_current_time()

    assert result1 != result2, "Cache did not expire as expected."

def test_cache_with_various_argument_types():
    cache = Cache(max_age_seconds=2, max_size=5)

    class Sut:
        @cache.cache
        def concat_elements(self, *args):
            return ''.join(map(str, args))

    sut = Sut()
    # Test with string arguments
    assert sut.concat_elements("hello", "world") == "helloworld"
    assert sut.concat_elements("hello", "world") == "helloworld"  # Cached result

    # Test with a mix of types
    assert sut.concat_elements("number", 123, (1, 2, 3)) == "number123(1, 2, 3)"
    assert sut.concat_elements("number", 123, (1, 2, 3)) == "number123(1, 2, 3)"  # Cached result

def test_cache_function_no_arguments():
    call_count = 0
    cache = Cache(max_age_seconds=2, max_size=5, enabled='enabled')

    class Sut:
        @cache.cache
        def get_fixed_value(self):
            nonlocal call_count
            call_count += 1
            return 42
        
    sut = Sut()

    assert sut.get_fixed_value() == 42  # First call, cache miss
    assert sut.get_fixed_value() == 42  # Second call, should hit cache
    assert call_count == 1, "Function was called more than once, indicating a cache miss"



from dataclasses import dataclass

from pydantic import BaseModel, field_validator
from enma.application.core.handlers.error import InvalidRequest
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.pagination import Pagination

class PaginateRequestDTO(BaseModel):
    page: int

    @field_validator("page")
    def validate_page(cls, page: int) -> int:
        if page <= 0:
            raise InvalidRequest(message='Page value must be greater than 0.')
        return int(page)
@dataclass
class PaginateResponseDTO:
    result: Pagination
    
class PaginateUseCase(IUseCase[PaginateRequestDTO, PaginateResponseDTO]):

    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self, dto: DTO[PaginateRequestDTO]) -> PaginateResponseDTO:
        logger.info(f'Retrieving page {dto.data.page}')
        result = self.__manga_repository.paginate(page=dto.data.page)
        
        return PaginateResponseDTO(result=result)



"""
This module defines the SearchResult entity for the Enma application.
It represents the result of a manga search operation.
"""
from dataclasses import dataclass
from datetime import datetime
from typing import TypedDict, Union
from enma.domain.entities.base import Entity
from enma.domain.entities.manga import Image, Manga

@dataclass
class Thumb:
    id: str
    url: str
    title: str
    cover: Image

class ISearchResultProps(TypedDict):
    id: Union[int, str]
    created_at: datetime
    updated_at: datetime
    page: int
    total_pages: int
    total_results: int
    results: list[Manga]

class Pagination(Entity[ISearchResultProps]):
    """
    Entity class representing a search result in the Enma application.
    """
    page: int
    total_pages: int
    total_results: int
    results: list[Thumb]

    def __init__(self,
                 page: int,
                 total_pages: int = 0,
                 total_results: int = 0,
                 results: Union[list[Thumb], None] = None,
                 id: Union[int, str, None] = None, 
                 created_at: Union[datetime, None] = None, 
                 updated_at: Union[datetime, None] = None):
        
        super().__init__(id=id,
                         created_at=created_at,
                         updated_at=updated_at)
        
        self.page = page
        self.results = results or list()
        self.total_pages = total_pages
        self.total_results = total_results

    def add_result(self, result: Thumb):
        self.results.append(result)


from dataclasses import dataclass

from pydantic import BaseModel
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.manga import Chapter, SymbolicLink

class FetchChapterBySymbolicLinkRequestDTO(BaseModel):
    link: SymbolicLink

@dataclass
class FetchChapterBySymbolicLinkResponseDTO:
    chapter: Chapter
    
class FetchChapterBySymbolicLinkUseCase(IUseCase[FetchChapterBySymbolicLinkRequestDTO, FetchChapterBySymbolicLinkResponseDTO]):

    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self, dto: DTO[FetchChapterBySymbolicLinkRequestDTO]) -> FetchChapterBySymbolicLinkResponseDTO:
        logger.info(f'Fetching chapter by symbolic link: {dto.data.link}.')
        
        chapter = self.__manga_repository.fetch_chapter_by_symbolic_link(link=dto.data.link)
        
        return FetchChapterBySymbolicLinkResponseDTO(chapter=chapter)



class SourceNotAvailable(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'SOURCE_NOT_AVAILABLE'
        self.desc: str = 'This error occurs when the client chooses nonexistent source.'
        self.critical: bool = False

class Unknown(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.code: str = 'UNKNOWN'
        self.desc: str = 'This error occours when was not possible to determine the error root cause.'
        self.critical: bool = True

class NotFound(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.code: str = 'NOT_FOUND'
        self.desc: str = 'This error occours when was not possible to find the requested resource.'
        self.critical: bool = True

class Forbidden(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.code: str = 'FORBIDDEN'
        self.desc: str = 'This error occours when the client can\'t perform a request to the source due lack of credentials.'
        self.critical: bool = True

class InvalidRequest(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.code: str = 'INVALID_REQUEST'
        self.desc: str = 'This error occours when the client tries to perform an request with wrong data input.'
        self.critical: bool = True

class InvalidResource(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'INVALID_RESOURCE'
        self.desc: str = 'This error occurs when the client tries to perform an action with an invalid resource.'
        self.critical: bool = True

class NhentaiSourceWithoutConfig(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'NHENTAI_SOURCE_WITHOUT_CONFIG'
        self.desc: str = 'This error occurs when the client tries to use Nhentai source with no cloudflare cookie.'
        self.critical: bool = True

class InvalidConfig(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'INVALID_CONFIG'
        self.desc: str = 'This error occurs when client tries to set wrong config object.'
        self.critical: bool = True

class InstanceError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'INSTANCE_ERROR'
        self.desc: str = 'This error occurs when the client tries to create an instance with wrong type.'
        self.critical: bool = True

class SourceWasNotDefined(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'SOURCE_WAS_NOT_DEFINED'
        self.desc: str = 'This error occurs when the client tries to perform Enma operations before of setting a source.'
        self.critical: bool = True

class ExceedRetryCount(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'EXCEED_RETRY_COUNT'
        self.desc: str = 'This error occurs when enma tries perform some action but something went wrong.'
        self.critical: bool = True

class ExceedRateLimit(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.message: str = message
        self.code: str = 'EXCEED_RATE_EXCEED'
        self.desc: str = 'This error occurs when enma perform more requests than a server can handle. Cool down your requests to this source!'
        self.critical: bool = False



from abc import ABC, abstractmethod
from io import BytesIO
from enma.domain.entities.manga import Image


class IDownloaderAdapter(ABC):

    @abstractmethod
    def download(self, page: Image) -> BytesIO:
        ...



import os
from expiringdict import ExpiringDict
from typing import Callable, Literal, Union, cast
from enma.application.core.utils.logger import logger

CacheState = Union[Literal["enabled"], Literal['disabled']]
STATE = cast(CacheState, os.getenv("ENMA_CACHING_STATE", 'enabled'))

class Cache:

  def __init__(self, 
               max_age_seconds: int=3600, 
               max_size: int=100, 
               enabled: CacheState = STATE):
    self._CACHE = ExpiringDict(max_len=max_size, 
                               max_age_seconds=max_age_seconds, 
                               items=None)
    self.enabled: bool = enabled == 'enabled'
  
  def cache(self, function: Callable):
    def wrapper(*args, **kwargs):

        _args = list()

        if len(kwargs.keys()) == 0:
            _args = list(args[1:])
        else:
            _args = list([*list(kwargs.values()), *args[1:]])

        if self._CACHE.get(str(_args)) is not None and self.enabled:
            logger.debug(f'Retrieving cached object with key {str(_args)}')
            return self._CACHE.get(str(_args))
        else:
            new_execution = function(*args, **kwargs)
            self._CACHE[str(_args)] = new_execution
            return new_execution

    return wrapper



from io import BytesIO
import random
import requests

from enma.application.core.interfaces.downloader_adapter import IDownloaderAdapter
from enma.domain.entities.manga import Image


class DefaultDownloader(IDownloaderAdapter):
    def download(self, page: Image) -> BytesIO:
        response = requests.get(url=page.uri)
        return BytesIO(response.content)



import sys
from enma.application.core.utils.logger import LogMode, logger
from enma.application.use_cases.download_chapter import Threaded
from enma.infra.entrypoints.lib import Enma, Sources
from enma.infra.adapters.repositories.nhentai import CloudFlareConfig, NHentai, Sort as NHentaiSort, NhentaiSourceConfig, NHentaiConfig
from enma.infra.adapters.repositories.mangadex import Mangadex, Sort as MangadexSort
from enma.infra.adapters.repositories.manganato import Manganato
from enma.infra.adapters.downloaders.default import DefaultDownloader
from enma.infra.adapters.downloaders.manganato import ManganatoDownloader
from enma.application.core.interfaces.downloader_adapter import IDownloaderAdapter
from enma.application.core.interfaces.saver_adapter import ISaverAdapter
from enma.infra.adapters.storage.google_drive import GoogleDriveStorage
from enma.infra.adapters.storage.local import LocalStorage
from enma.domain.entities.manga import Manga, Chapter, SymbolicLink
from enma.domain.entities.search_result import SearchResult
from enma.domain.entities.pagination import Pagination
from enma.domain.entities.author_page import AuthorPage
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma._version import __version__

package_name = "enma"
python_major = "3"
python_minor = "9"

logger.mode = LogMode.SILENT

try:
    assert sys.version_info >= (int(python_major), int(python_minor))
except AssertionError:
    raise RuntimeError(f"\033[31m{package_name!r} requires Python {python_major}.{python_minor}+ (You have Python {sys.version})\033[0m")



from dataclasses import dataclass
from typing import Any
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.manga import Manga

@dataclass
class RandomResponseDTO:
    result: Manga
    
class RandomUseCase(IUseCase[Any, RandomResponseDTO]):

    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self) -> RandomResponseDTO:
        logger.info(f'Fetching random manga.')
        result = self.__manga_repository.random()
        
        return RandomResponseDTO(result=result)



import copy
from unittest.mock import MagicMock, Mock, patch
import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.infra.core.interfaces.nhentai_response import NHentaiImage
from enma.application.core.handlers.error import Forbidden, InvalidConfig, InvalidRequest, NhentaiSourceWithoutConfig, NotFound
from enma.domain.entities.pagination import Thumb
from tests.data.mocked_doujins import (nhentai_doujin_mocked, 
                                       nhentai_search_mocked, 
                                       nhentai_search_no_result_mocked, 
                                       nhentai_paginate_mocked, 
                                       nhentai_author_mocked,
                                       nhentai_author_not_found_mocked)
from enma.infra.adapters.repositories.nhentai import CloudFlareConfig, NHentai, NHentaiConfig, Sort
from enma.domain.entities.manga import MIME, Author, Chapter, Genre, Image, SymbolicLink
from enma.application.core.utils.logger import logger

class TestNHentaiUtils:
    sut = NHentai(
        config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        )
    )
    
    def test_raise_error_if_passing_wrong_config(self):
        with pytest.raises(InvalidConfig) as err:
            self.sut.set_config(config={}) # type: ignore
    
    def test_set_config_successfully(self):
        res = self.sut.set_config(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))
        assert res is None

    def test_set_config_must_raise_in_case_of_empty_values(self):
        with pytest.raises(InvalidRequest):
            self.sut.set_config(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='',
                cf_clearance=''
            )
        ))

    def test_should_run_normally_even_without_config(self):
        self.sut._NHentai__config = None # type: ignore
        req = self.sut._NHentai__make_request(url='https://www.google.com') # type: ignore
        assert req.status_code == 200

    @patch('requests.get')
    def test_make_a_request_successfully(self, request_mock: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = {}
        request_mock.return_value = mock

        self.sut.set_config(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))
        res = self.sut._NHentai__make_request(url='https://www.google.com', headers={'add': 'header'}, params={'new': 'parameter'}) # type: ignore

        assert res.status_code == 200
        assert res.json() == {}

        request_mock.assert_called_once_with(url='https://www.google.com', 
                                             headers={'User-Agent': 'mocked', 
                                                      'add': 'header'},
                                             params={'new': 'parameter'},
                                             cookies={'cf_clearance': 'mocked'})
        
    def test_should_filter_tags_correctly(self):
        TAGS = [
            {
                "id": 33172,
                "type": "category",
                "name": "doujinshi",
                "url": "/category/doujinshi/",
                "count": 228729
            },
            {
                "id": 29859,
                "type": "tag",
                "name": "b",
                "url": "/tag/b/",
                "count": 37321
            },
            {
                "id": 1,
                "type": "tag",
                "name": "a",
                "url": "/tag/a/",
                "count": 1
            }
        ]

        characters = self.sut._NHentai__get_tag_by_type(type='category', tags=TAGS) # type: ignore
        tags = self.sut._NHentai__get_tag_by_type(type='tag', tags=TAGS) # type: ignore
        assert len(characters) == 1
        assert len(tags) == 2

    def test_making_page_uri(self):
        page = self.sut._NHentai__make_page_uri(type='page', media_id='1234', mime=MIME.J, page_number=1) # type: ignore
        assert page == 'https://i1.nhentai.net/galleries/1234/1.jpg'

    def test_making_cover_uri(self):
        page = self.sut._NHentai__make_page_uri(type='cover', media_id='1234', mime=MIME.J, page_number=1) # type: ignore
        assert page == 'https://t1.nhentai.net/galleries/1234/cover.jpg'

    def test_making_thumbnail_uri(self):
        page = self.sut._NHentai__make_page_uri(type='thumbnail', media_id='1234', mime=MIME.J, page_number=1) # type: ignore
        assert page == 'https://t1.nhentai.net/galleries/1234/thumb.jpg'

    def test_chapter_creator(self):
        images: list[NHentaiImage] = [{
            'h': 123,
            'w': 123,
            't': 'j'
        }]
        chapter = self.sut._NHentai__create_chapter(url='https://nhentai.net/g/1234', with_symbolic_links=False, media_id='12', pages=images) # type: ignore
        assert isinstance(chapter, Chapter)
        assert len(chapter.pages) == 1
        assert chapter.pages_count == 1
        assert isinstance(chapter.pages[0], Image)
        assert chapter.pages[0].width == 123
        assert chapter.pages[0].height == 123
        assert chapter.pages[0].mime.value == 'jpg'
        assert chapter.link is None

    def test_chapter_creator_with_symbolic_links(self):
        images: list[NHentaiImage] = [{
            'h': 123,
            'w': 123,
            't': 'j'
        }]
        chapter = self.sut._NHentai__create_chapter(url='https://nhentai.net/g/1234', with_symbolic_links=True, media_id='12', pages=images) # type: ignore
        assert isinstance(chapter, Chapter)
        assert len(chapter.pages) == 0
        assert chapter.pages_count == 0
        assert chapter.link == SymbolicLink(link='https://nhentai.net/g/1234')

class TestNHentaiSourceGetMethod:

    sut = NHentai(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))

    @patch('requests.get')
    def test_success_doujin_retrieve(self, sut_mock: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = nhentai_doujin_mocked

        sut_mock.return_value = mock

        res = self.sut.get(identifier='1')

        assert res is not None
        assert res.id == 1
        assert res.title.english == "(C71) [Arisan-Antenna (Koari)] Eat The Rich! (Sukatto Golf Pangya)"
        assert res.title.japanese == "(C71) [ありさんアンテナ (小蟻)] Eat The Rich! (スカッとゴルフ パンヤ)"
        assert res.title.other == "Eat The Rich!"
        
        for genre in res.genres:
            assert isinstance(genre, Genre)
        
        for author in res.authors:
            assert isinstance(author, Author)

        assert isinstance(res.thumbnail, Image)
        assert isinstance(res.cover, Image)

        for chapter in res.chapters:
            assert isinstance(chapter, Chapter)
        
        assert len(res.tags) != 0

    @patch('requests.get')
    def test_must_return_other_titles_as_none_if_doesnt_exists(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200

        data = copy.deepcopy(nhentai_doujin_mocked)    
        data['title']['japanese'] = None # type: ignore
        data['title']['pretty'] = None # type: ignore

        mock.json.return_value = data
        
        mock_method.return_value = mock
        
        manga = self.sut.get(identifier='489504')

        assert manga is not None
        assert manga.id == 1
        assert manga.title.english == "(C71) [Arisan-Antenna (Koari)] Eat The Rich! (Sukatto Golf Pangya)"
        assert manga.title.japanese == None
        assert manga.title.other == None

    @patch('requests.get')
    def test_response_when_it_could_not_get_doujin(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 404
        mock_method.return_value = mock

        with pytest.raises(NotFound):
            self.sut.get(identifier='1')
    
    @patch('requests.get')
    def test_raise_forbidden_when_receive_403_status_code(self, mock_method: MagicMock):
            mock = Mock()
            mock.status_code = 403
            mock_method.return_value = mock

            with pytest.raises(Forbidden):
                self.sut.get(identifier='1')
                mock_method.assert_called_with(url=f'https://nhentai.net/api//gallery/1',
                                            headers={'User-Agent': 'mock'},
                                            params={},
                                            cookies={'cf_clearance': 'mock'})
    
    @patch('requests.get')
    def test_return_empty_chapters(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        
        data = copy.deepcopy(nhentai_doujin_mocked) 
        data['images']['pages'] = []
        mock.json.return_value = data

        mock_method.return_value = mock

        doujin = self.sut.get(identifier='1')

        assert doujin is not None
        assert isinstance(doujin.chapters[0], Chapter)
        assert len(doujin.chapters[0].pages) == 0
        assert doujin.id == 1 

    @patch('requests.get')
    def test_return_right_mime(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200

        mock.json.return_value = nhentai_doujin_mocked

        mock_method.return_value = mock

        doujin = self.sut.get(identifier='1')

        assert doujin is not None
        assert isinstance(doujin.chapters[0], Chapter)
        assert doujin.chapters[0].pages[0].mime.value == 'jpg'
        assert nhentai_doujin_mocked['images']['pages'][0]['t'] == 'j'
        assert doujin.cover is not None
        assert doujin.cover.mime.value == 'png'
        assert nhentai_doujin_mocked['images']['cover']['t'] == 'p'
        assert doujin.chapters[0].link is None
    
    @patch('requests.get')
    def test_get_with_symbolic_link(self, mock_method: MagicMock):
            mock = Mock()
            mock.status_code = 200

            mock.json.return_value = nhentai_doujin_mocked

            mock_method.return_value = mock

            doujin = self.sut.get(identifier='420719', with_symbolic_links=True)

            assert doujin is not None
            assert isinstance(doujin.chapters[0], Chapter)
            assert doujin.chapters[0].link is not None
            assert doujin.chapters[0].link != ""

    @patch('requests.get')
    def test_language_must_be_present(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = nhentai_doujin_mocked

        mock_method.return_value = mock

        doujin = self.sut.get(identifier='420719', with_symbolic_links=True)

        assert doujin is not None
        assert doujin.language is not None
        assert doujin.language == 'japanese'

    @patch('requests.get')
    def test_images_mime_types_must_be_correct(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = nhentai_doujin_mocked

        mock_method.return_value = mock

        doujin = self.sut.get(identifier='420719', with_symbolic_links=True)

        assert doujin is not None
        assert doujin.cover is not None        
        assert doujin.thumbnail is not None

        cover_mime = nhentai_doujin_mocked['images']['cover']['t']
        thumb_mime = nhentai_doujin_mocked['images']['thumbnail']['t']

        assert doujin.thumbnail is not None
        assert doujin.cover is not None
        assert cover_mime.upper() == 'P'
        assert doujin.cover.mime.value == 'png'
        assert thumb_mime.upper() == 'J'
        assert doujin.thumbnail.mime.value == 'jpg'

class TestNHentaiSourcePaginationMethod:
    sut = NHentai(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))

    @patch('requests.get')
    def test_success_searching(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = nhentai_paginate_mocked

        mock_method.return_value = mock

        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == nhentai_paginate_mocked['num_pages']
        assert res.total_results == 25 * 19163
        assert len(res.results) == nhentai_paginate_mocked['per_page']
        
    @patch('requests.get')
    def test_must_return_empty_search_result(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200

        copied = copy.deepcopy(nhentai_paginate_mocked)
        copied['result'] = [] # type: ignore        
        mock.json.return_value = copied

        mock_method.return_value = mock
        
        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == nhentai_paginate_mocked['num_pages']
        assert res.total_results == 25 * 19163
        assert len(res.results) == 0

class TestNHentaiSourceSearchMethod:

    sut = NHentai(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))

    @patch('requests.get')
    def test_success_searching(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.text = nhentai_search_mocked

        mock_method.return_value = mock

        res = self.sut.search(query='GATE', page=2)

        assert res is not None
        assert res.query == 'GATE'
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == 3
        assert len(res.results) == 25

        for result in res.results:
            assert isinstance(result, Thumb)
            assert result.cover is not None
            assert isinstance(result.cover, Image)
            assert isinstance(result.cover.width, int)
            assert isinstance(result.cover.height, int)
            assert result.cover.width > 0
            assert result.cover.height > 0
            assert result.title is not None
            assert result.id != 0

        mock_method.assert_called_once_with(url='https://nhentai.net/search',
                                            params={'q': 'GATE',
                                                    'sort': Sort.RECENT.value,
                                                    'page': 2},
                                            headers={'User-Agent': 'mocked'},
                                            cookies={'cf_clearance': 'mocked'})
        
    @patch('requests.get')
    def test_must_return_empty_search_result(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.text = nhentai_search_no_result_mocked
        mock_method.return_value = mock
        
        res = self.sut.search(query='GATE', page=2)

        assert res is not None
        assert res.query == 'GATE'
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == 0
        assert len(res.results) == 0


class TestNHentaiSourceAuthorPageMethod:

    sut = NHentai(config=NHentaiConfig(
            cloudflare=CloudFlareConfig(
                user_agent='mocked',
                cf_clearance='mocked'
            )
        ))

    @patch('requests.get')
    def test_success_author_page_fetching(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.text = nhentai_author_mocked

        mock_method.return_value = mock

        res = self.sut.author_page(author='akaneman', page=2)

        assert res is not None
        assert res.author == 'akaneman'
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == 2
        assert len(res.results) == 25

        for result in res.results:
            assert isinstance(result, Thumb)
            assert result.cover is not None
            assert isinstance(result.cover, Image)
            assert isinstance(result.cover.width, int)
            assert isinstance(result.cover.height, int)
            assert result.cover.width > 0
            assert result.cover.height > 0
            assert result.title is not None
            assert result.id != 0

        mock_method.assert_called_once_with(url='https://nhentai.net/artist/akaneman',
                                            params={'page': 2},
                                            headers={'User-Agent': 'mocked'},
                                            cookies={'cf_clearance': 'mocked'})
        
    @patch('requests.get')
    def test_must_return_empty_author_page_result(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.text = nhentai_author_not_found_mocked
        mock_method.return_value = mock
        
        res = self.sut.author_page(author='asdsadadasd', page=1)

        assert res is not None
        assert res.author == 'asdsadadasd'
        assert res.id is not None
        assert res.page == 1
        assert res.total_pages == 0
        assert len(res.results) == 0




from dataclasses import dataclass
from typing import Union

from pydantic import BaseModel, Field, validator
from enma.application.core.handlers import error
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.manga import Manga

class GetMangaRequestDTO(BaseModel):
    identifier: str
    with_symbolic_links: bool = Field(default=False)

@dataclass
class GetMangaResponseDTO:
    found: bool
    manga: Union[Manga, None]
class GetMangaUseCase(IUseCase[GetMangaRequestDTO, GetMangaResponseDTO]):
    
    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self, dto: DTO[GetMangaRequestDTO]) -> GetMangaResponseDTO:
        logger.info(f'Fetching manga with identifier: {dto.data.identifier}.')

        try:
            manga = self.__manga_repository.get(identifier=dto.data.identifier,
                                                with_symbolic_links=dto.data.with_symbolic_links)
            return GetMangaResponseDTO(found=True, manga=manga)
        except error.NotFound:
            logger.error(f'Could not find the manga using the provided identifier: {dto.data.identifier}')
            return GetMangaResponseDTO(found=False, manga=None)



import json

from enma.infra.core.interfaces.mangadex_response import ISearchResult
from enma.infra.core.interfaces.nhentai_response import NHentaiPaginateResponse, NHentaiResponse

with open('./tests/data/get.json', 'r') as get:
    nhentai_doujin_mocked: NHentaiResponse = json.loads(get.read())

with open('./tests/data/nhentai_search_mock.txt', 'r') as search:
    nhentai_search_mocked = search.read()

with open('./tests/data/nhentai_search_no_result.txt', 'r') as no_result:
    nhentai_search_no_result_mocked = no_result.read()

with open('./tests/data/nhentai_paginate_mock.json', 'r') as paginate:
    nhentai_paginate_mocked: NHentaiPaginateResponse = json.loads(paginate.read())

with open('./tests/data/nhentai_author_page_mock.txt', 'r') as author:
    nhentai_author_mocked = author.read()

with open('./tests/data/nhentai_author_page_not_found_mock.txt', 'r') as author_not_found:
    nhentai_author_not_found_mocked = author_not_found.read()

with open('./tests/data/manganato_manga_page_empty_chapters_mock.txt', 'r') as empty_ch:
    manganato_manga_page_empty_chapters_mocked = empty_ch.read()

with open('./tests/data/manganato_manga_page_empty_pagination_mock.txt', 'r') as empty_pag:
    manganato_manga_page_empty_pagination_mocked = empty_pag.read()

with open('./tests/data/mangadex_empty_pagination_mock.json', 'r') as paginate:
    mangadex_paginate_mocked: ISearchResult = json.loads(paginate.read())



from dataclasses import dataclass, field
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.author_page import AuthorPage
from pydantic import BaseModel, Field

class GetAuthorPageRequestDTO(BaseModel):
    author: str
    page: int = Field(default=1)

@dataclass
class GetAuthorPageResponseDTO:
    result: AuthorPage
    
class GetAuthorPageUseCase(IUseCase[GetAuthorPageRequestDTO, GetAuthorPageResponseDTO]):

    def __init__(self, manga_repository: IMangaRepository):
        self.__manga_repository = manga_repository

    def execute(self, dto: DTO[GetAuthorPageRequestDTO]) -> GetAuthorPageResponseDTO:
        logger.info(f'Retrieving {dto.data.author} page.')
        result = self.__manga_repository.author_page(author=dto.data.author,
                                                     page=dto.data.page)
        
        return GetAuthorPageResponseDTO(result=result)



"""
This module provides an adapter for the mangadex repository.
It contains functions and classes to interact with the mangadex API and retrieve manga data.
"""
from datetime import datetime
from enum import Enum
import os
from typing import Any, Optional, Union, cast
from urllib.parse import urljoin, urlparse

import requests

from enma._version import __version__
from enma.application.core.handlers.error import (ExceedRateLimit, 
                                                  Forbidden, 
                                                  NotFound, 
                                                  Unknown)
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.utils.logger import logger
from enma.domain.entities.author_page import AuthorPage
from enma.domain.entities.manga import (MIME, 
                                        Chapter, 
                                        Genre, 
                                        Author, 
                                        Image, 
                                        Language, 
                                        Manga, 
                                        SymbolicLink,
                                        Title)
from enma.domain.entities.search_result import Pagination, SearchResult, Thumb
from enma.domain.utils import mime
from enma.infra.core.interfaces.mangadex_response import (AuthorRelation, 
                                                          CoverArtRelation, IAltTitles, 
                                                          IGetResult, 
                                                          IHash, IManga, IMangaTag, IRelations, 
                                                          ISearchResult, 
                                                          IVolumesResponse)
from enma.infra.core.utils.cache import Cache


class Sort(Enum):
    ALL_TIME = 'relevance'
    RECENT = 'createdAt'

class Mangadex(IMangaRepository):
    """
    Repository class for interacting with the Mangadex API.
    Provides methods to fetch manga details, search for manga, etc.
    """

    def __init__(self) -> None:
        self.__SITE_URL = 'https://mangadex.org/'
        self.__API_URL = 'https://api.mangadex.org/'
        self.__COVER_URL = 'https://mangadex.org/covers/'
        self.__HASH_URL = 'https://api.mangadex.org/at-home/server/'
        self.__CHAPTER_PAGE_URL = 'https://cmdxd98sb0x3yprd.mangadex.network/data/'

    def __handle_source_response(self, response: requests.Response):
        """
        Evaluates the HTTP response from the Mangadex API, raising specific exceptions based on the HTTP status code
        to indicate various error conditions such as rate limits exceeded, forbidden access, or resource not found.

        Args:
            response (Response): The HTTP response object from a request to the Mangadex API.

        Raises:
            Forbidden: Indicates a 403 Forbidden HTTP status code.
            NotFound: Indicates a 404 Not Found HTTP status code.
            ExceedRateLimit: Indicates a 429 Too Many Requests HTTP status code.
            Unknown: Indicates any other unexpected HTTP status code.
        """

        logger.debug(f'Fetched {response.url} with response status code {response.status_code} and text {response.text}')

        if response.status_code == 200:
            return
        if response.status_code == 403:
            raise Forbidden(message='Could not perform a successful request to the source due to credentials issues. Check your credentials and try again.')
        if response.status_code == 404:
            raise NotFound(message=f'Could not find the requested resource at "{response.url}". Check the provided request parameters and try again.')
        if response.status_code == 429:
            raise ExceedRateLimit(message='You have exceeded the Mangadex rate limit!')
        raise Unknown(message='Something unexpected happened while trying to fetch source content. Set the logging mode to debug and try again.')

    def __make_request(self,
                       url: str,
                       headers: Union[dict[str, Any], None] = None,
                       params: Optional[Union[dict[str, Union[str, int]], list[tuple[str, Union[str, int]]]]] = None) -> requests.Response:
        """
        Makes a request to the specified URL with the given headers and parameters.

        Args:
            url (str): The URL to make the request to.
            headers (dict[str, Any], optional): The headers to include in the request. Defaults to None.
            params (Optional[Union[dict[str, Union[str, int]], list[tuple[str, Union[str, int]]]]], optional): The parameters to include in the request. Defaults to None.

        Returns:
            requests.Response: The response object from the API request.
        """
        headers = headers if headers is not None else {}
        params = params if params is not None else {}

        logger.debug(f'Fetching {url} with headers {headers} and params {params}')

        response = requests.get(url=urlparse(url).geturl(),
                                headers={**headers, "User-Agent": f"Enma/{__version__}"},
                                params=params)
        
        self.__handle_source_response(response)

        return response

    def set_config(self, config) -> None:
        raise NotImplementedError('Manganato does not support set config')

    def __create_cover_uri(self, 
                           manga_id: str, 
                           file_name: str) -> str:
        """
        Constructs a URL for a manga's cover image based on its identifier and the file name of the cover image.

        Args:
            manga_id (str): The unique identifier of the manga.
            file_name (str): The file name of the cover image.

        Returns:
            str: The fully qualified URL to the cover image.
        """
        return urljoin(self.__COVER_URL, f'{manga_id}/{file_name}.512.jpg')
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_FETCH_SYMBOLIC_LINK_TTL_IN_SECONDS', 100)), 
           max_size=20).cache
    def fetch_chapter_by_symbolic_link(self, 
                                       link: SymbolicLink) -> Chapter:
        """
        Retrieves manga chapter details including pages and images by following a symbolic link. This method is particularly
        useful for fetching chapters that have been directly linked.

        Args:
            link (SymbolicLink): An object representing the symbolic link to the chapter.

        Returns:
            Chapter: An object containing the fetched chapter details such as pages and images.
        """
        response = self.__make_request(url=link.link)
        
        ch: IHash = response.json()
        chapter = Chapter()

    
        for index, page in enumerate(ch.get('chapter').get('data')):
            extension = page.split('.')[-1]
            safe_mime = mime.get_mime_safelly(extension.upper())
            
            if safe_mime is None:
                logger.warning(f'Could not determine MIME type for extension {extension}. Defaulting to JPEG.')

            safe_mime = safe_mime if safe_mime is not None else MIME.J

            chapter.add_page(Image(uri=self.__create_chapter_page_uri(ch.get('chapter').get('hash'), page),
                                   name=f'{index}.{extension}',
                                   mime=safe_mime))
        return chapter

    def __fetch_chapter_hashes(self, chapter_id: str) -> tuple[str, list[str]]:
        """
        Fetches the chapter hashes and page file names for a given chapter ID. These details are necessary to construct
        the URLs for individual chapter pages.

        Args:
            chapter_id (str): The unique identifier of the chapter.

        Returns:
            tuple[str, list[str]]: A tuple containing the chapter hash and a list of page file names.
        """
        response = self.__make_request(url=urljoin(self.__HASH_URL, chapter_id))
        data: IHash = response.json()

        return (data.get('chapter').get('hash'), 
                data.get('chapter').get('data'))
    
    def __create_chapter_page_uri(self, hash: str, filename: str) -> str:
        """
        Constructs the URL for a chapter page given the chapter hash and the page file name.

        Args:
            hash (str): The hash of the chapter, used as part of the URL path.
            filename (str): The file name of the chapter page.

        Returns:
            str: The fully qualified URL to the chapter page.
        """
        return urljoin(self.__CHAPTER_PAGE_URL, f'{hash}/{filename}')

    def __create_chapter(self,
                         chapter: tuple[int, str],
                         with_symbolic_links: bool = False) -> Chapter:
        """
        Constructs a Chapter object for a given chapter tuple, optionally using symbolic links. If symbolic links are
        used, chapter pages are not pre-fetched but are instead represented as links.

        Args:
            chapter (tuple[int, str]): A tuple containing the chapter number and the chapter ID.
            with_symbolic_links (bool, optional): A flag indicating whether to use symbolic links for chapter pages. Defaults to False.

        Returns:
            Chapter: The constructed Chapter object.
        """        
        curr_chapter, chapter_id = chapter

        if with_symbolic_links:
            return Chapter(id=curr_chapter, 
                           link=SymbolicLink(link=urljoin(self.__HASH_URL, chapter_id)))
        else:
            ch = Chapter(id=curr_chapter)
            hash, files = self.__fetch_chapter_hashes(chapter_id)
        
            for index, page in enumerate(files):
                extension = page.split('.')[-1]
                safe_mime = mime.get_mime_safelly(extension.upper())
            
                if safe_mime is None:
                    logger.warning(f'Could not determine MIME type for extension {extension}. Defaulting to JPEG.')

                safe_mime = safe_mime if safe_mime is not None else MIME.J
                ch.add_page(Image(uri=self.__create_chapter_page_uri(hash, page),
                                  name=f'{index}.{extension}',
                                  mime=safe_mime))
        
            return ch
    
    def __list_chapters(self, manga_id: str) -> list[tuple[int, str]]:
        """
        Retrieves a list of chapters for a given manga ID. Each chapter is represented as a tuple containing the chapter
        number and the chapter ID.

        Args:
            manga_id (str): The unique identifier of the manga.

        Returns:
            list[tuple[int, str]]: A list of tuples, each representing a chapter of the manga.
        """
        response = self.__make_request(url=urljoin(self.__API_URL, f'manga/{manga_id}/aggregate'))
        
        data: IVolumesResponse = response.json()

        chapters = list()
        for volume in data.get('volumes', dict()):
            volume = data.get('volumes').get(volume)
            
            if volume is None: continue

            volume_chapters = volume.get('chapters', dict())
            
            for volume_key in volume_chapters:
                current_vol = volume_chapters.get(volume_key)

                if current_vol is None: continue

                chapters.append((current_vol.get('chapter'), current_vol.get('id')))
        
        return chapters
    
    def __extract_authors(self, relations: IRelations) -> list[Author]:
        """
        Extracts author information from a list of relationships within manga metadata, constructing Author objects
        for each author found.

        Args:
            relations (IRelations): A list of relationship objects from the manga metadata.

        Returns:
            list[Author]: A list of Author objects extracted from the relationships.
        """
        authors_data = [relationship for relationship in relations if relationship.get('type') == 'author']
        authors: list[Author] = []

        if len(authors_data) > 0:
            for author in authors_data:
                author = cast(AuthorRelation, author)
                authors.append(Author(name=author.get('attributes').get('name'),
                                      id=author.get('id')))
                
        return authors
    
    def __extract_genres(self, tags: list[IMangaTag]) -> list[Genre]:
        """
        Extracts genre information from a list of tags within manga metadata, constructing Genre objects for each tag
        that represents a genre.

        Args:
            tags (list[IMangaTag]): A list of tag objects from the manga metadata.

        Returns:
            list[Genre]: A list of Genre objects extracted from the tags.
        """
        return [Genre(id=tag.get('id'),
                      name=tag.get('attributes', {}).get('name', {}).get('en', 'unknown')) 
                    for tag in tags or []
                    if tag.get('type') == 'tag']
    
    def __get_cover(self, 
                    manga_id: str, 
                    relations: IRelations) -> Image:
        """
        Retrieves the cover image for a given manga ID from a list of relationships. If a cover image is found, an
        Image object is constructed and returned.

        Args:
            manga_id (str): The unique identifier of the manga.
            relations (IRelations): A list of relationship objects from the manga metadata.

        Returns:
            Image: An Image object representing the manga's cover image. Returns an Image object with an empty URI if no cover is found.
        """
        covers = [tag for tag in relations if tag.get('type') == 'cover_art']
        if len(covers) == 0: return Image(uri='')
        cover = cast(CoverArtRelation, covers[0])
        return Image(uri=self.__create_cover_uri(manga_id, cover.get("attributes").get("fileName")),
                     width=512)
    
    def __get_title(self, alt_titles: IAltTitles, title: str) -> Title:
        """
        Constructs a Title object for the manga, incorporating the English title, a Japanese title if available,
        and an alternative title.

        Args:
            alt_titles (IAltTitles): A list of alternative titles for the manga.
            title (str): The primary English title of the manga.

        Returns:
            Title: A Title object containing the English, Japanese, and an alternative title for the manga.
        """
        japanese_titles = [ title.get('ja-ro') for title in alt_titles if title.get('ja-ro') is not None ]
        japanese_title = japanese_titles[0] if len(japanese_titles) > 0 else None

        other_keys = list(alt_titles[-1].keys()) if len(alt_titles) > 0 else []
        other_key = other_keys[0] if len(other_keys) > 0 else ''

        return Title(english=title,
                     japanese=japanese_title or '',
                     other=alt_titles[-1].get(other_key, '') if len(alt_titles) > 0 else '')

    def __parse_full_manga(self, 
                           manga_data: IManga, 
                           with_symbolic_links: bool = False) -> Manga:
        """
        Parses the complete manga data retrieved from the Mangadex API, constructing a Manga object that includes
        details such as title, authors, genres, cover image, and chapters.

        Args:
            manga_data (IManga): The raw manga data from the Mangadex API.
            with_symbolic_links (bool, optional): Indicates whether to use symbolic links for chapters. Defaults to False.

        Returns:
            Manga: A fully constructed Manga object.
        """
        attrs = manga_data.get('attributes', dict())

        thumbnail = self.__get_cover(manga_data.get('id'), 
                                     manga_data.get('relationships'))
        
        status = 'completed' if attrs.get('status', "").lower() == 'completed' else 'ongoing'

        manga = Manga(title=self.__get_title(alt_titles=attrs.get('altTitles'),
                                             title=attrs.get('title', dict()).get('en') or ''),
                      id=manga_data.get('id'),
                      created_at=datetime.fromisoformat(attrs.get('createdAt')),
                      updated_at=datetime.fromisoformat(attrs.get('updatedAt')),
                      status=status,
                      url=urljoin(self.__SITE_URL, f'title/{manga_data.get("id")}'),
                      language=Language.get(attrs.get('originalLanguage').strip().lower().replace('-', '_'), 'unknown'),
                      authors=self.__extract_authors(manga_data.get('relationships', list())),
                      genres=self.__extract_genres(attrs.get('tags', list())),
                      thumbnail=thumbnail,
                      cover=thumbnail)
        
        chapter_list = self.__list_chapters(manga_id=str(manga.id))

        for chapter in chapter_list:
            manga.add_chapter(self.__create_chapter(chapter=chapter,
                                                    with_symbolic_links=with_symbolic_links))
            
        return manga
    
    def __parse_thumb(self, manga: IManga) -> Thumb:
        """
        Extracts minimal manga information to construct a Thumb object, primarily used for search results
        where detailed information is not necessary.

        Args:
            manga (IManga): The raw manga data from the Mangadex API.

        Returns:
            Thumb: A Thumb object containing the manga's ID, title, and cover image.
        """

        title = manga.get('attributes').get('title').get('en')
        return Thumb(id=manga.get('id'),
                     title=title,
                     url=urljoin(self.__SITE_URL, f'title/{manga.get("id")}'),
                     cover=self.__get_cover(manga_id=manga.get('id'),
                                            relations=manga.get('relationships', list())))
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_GET_TTL_IN_SECONDS', 300)), 
           max_size=20).cache
    def get(self, 
            identifier: str,
            with_symbolic_links: bool = False) -> Manga:
        """
        Retrieves detailed information for a specific manga identified by its ID, constructing a Manga object.

        Args:
            identifier (str): The unique identifier of the manga to retrieve.
            with_symbolic_links (bool, optional): Indicates whether to construct the Manga object with symbolic links for chapters. Defaults to False.

        Returns:
            Manga: The Manga object containing detailed information about the specified manga.
        """
        response = self.__make_request(url=urljoin(self.__API_URL, f'manga/{identifier}'),
                                       params=[('includes[]', 'cover_art'),
                                               ('includes[]', 'author'),
                                               ('includes[]', 'artist')])

        result: IGetResult = response.json()
        manga_data = result.get('data')

        return self.__parse_full_manga(manga_data=manga_data,
                                       with_symbolic_links=with_symbolic_links)
        
    def __make_sort_query(self, sort: Sort) -> dict[str, str]:
        """
        Constructs a query parameter dictionary to define the sorting order for search results based on a Sort enumeration value.

        Args:
            sort (Sort): An enumeration value specifying the desired sort order for search results.

        Returns:
            dict[str, str]: A dictionary of query parameters to define the sorting order.
        """
        return { f'order[{sort.value if isinstance(sort, Sort) else sort}]': 'desc' }

    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_SEARCH_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def search(self,
               query: str,
               page: int,
               sort: Sort = Sort.RECENT,
               per_page: int = 25) -> SearchResult:
        """
        Searches the Mangadex API for manga that match the given query string, optionally sorting the results
        and paginating them. Constructs and returns a SearchResult object containing the search results.
        
        Args:
            query (str): The search query string.
            page (int): The page number of the search results to retrieve.
            sort (Sort, optional): The sorting order of the search results. Defaults to Sort.RECENT.
            per_page (int, optional): The number of results per page. Defaults to 25.

        Returns:
            SearchResult: An object containing the paginated search results, including manga thumbnails.
        """
        logger.debug(f'Searching into Mangadex with args query={query};page={page};sort={sort}')

        params = [('title', query), *tuple(self.__make_sort_query(sort).items()),
                  ('includes[]', 'cover_art'), ('limit', per_page),
                  ('offset', per_page * (page - 1) if page > 1 else 0), ('contentRating[]', 'safe'),
                  ('contentRating[]', 'suggestive'), ('contentRating[]', 'erotica'),
                  ('order[createdAt]', 'desc'), ('hasAvailableChapters', 'true')]
        
        request_response = self.__make_request(url=urljoin(self.__API_URL, 'manga'),
                                               params=params)
        
        response: ISearchResult = request_response.json()

        total_results = response.get('total')
        total_pages = int(total_results / per_page)
        
        search_result = SearchResult(query=query,
                                     total_pages=total_pages,
                                     page=page,
                                     total_results=total_results)
        
        for result in response.get('data', []):
            search_result.results.append(self.__parse_thumb(manga=result))

        return search_result

    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_PAGINATE_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def paginate(self, page: int) -> Pagination:
        """
        Retrieves a specific page of manga listings from the Mangadex API, returning a Pagination object
        that includes a list of manga thumbnails for that page.

        Args:
            page (int): The page number of manga listings to retrieve.

        Returns:
            Pagination: An object containing the paginated list of manga thumbnails and pagination details.
        """
        logger.debug(f'Paginating with args page={page}')
        per_page = 25
        request_response = self.__make_request(url=urljoin(self.__API_URL, 'manga'),
                                               params=[('limit', per_page),
                                                       ('offset', per_page * (page - 1) if page > 1 else 0),
                                                       ('order[createdAt]', 'desc'),
                                                       ('includes[]', 'cover_art'),
                                                       ('contentRating[]', 'safe'),
                                                       ('contentRating[]', 'suggestive'),
                                                       ('contentRating[]', 'erotica'),
                                                       ('order[createdAt]', 'desc'),
                                                       ('hasAvailableChapters', 'true')])
        
        response: ISearchResult = request_response.json()

        pagination = Pagination(page=page,
                                total_pages=int(response.get('total') / per_page),
                                total_results=response.get('total'))

        for result in response.get('data', []):            
            pagination.results.append(self.__parse_thumb(manga=result))

        return pagination

    def random(self, retry=0) -> Manga:
        """
        Fetches a random manga from the Mangadex API. If the first attempt fails, it will retry up to a specified number of times.

        Args:
            retry (int, optional): The number of retries to attempt in case of failure. Defaults to 0.

        Returns:
            Manga: A Manga object for the randomly selected manga.
        """
        response = self.__make_request(url=urljoin(self.__API_URL, f'manga/random'),
                                       params=[('includes[]', 'cover_art'),
                                               ('contentRating[]', 'safe'),
                                               ('contentRating[]', 'suggestive'),
                                               ('contentRating[]', 'erotica'),
                                               ('includes[]', 'author'),
                                               ('includes[]', 'artist')])

        result: IGetResult = response.json()

        manga = result.get('data')

        return self.__parse_full_manga(manga_data=manga,
                                       with_symbolic_links=True)
    
    def author_page(self,
                    author: str,
                    page: int) -> AuthorPage:
        """
        Fetches manga authored by a specific author. This method is not currently implemented for Mangadex
        and serves as a placeholder for potential future functionality.

        Args:
            author (str): The name or identifier of the author.
            page (int): The page number of results to retrieve.

        Raises:
            NotImplementedError: Indicates that this method is not supported or implemented.

        Returns:
            AuthorPage: An object containing a list of manga by the specified author. This is currently not implemented.
        """
        raise NotImplementedError('Mangadex does not support author page.')
